from enum import StrEnum
from dataclasses import dataclass

from toydl.core.module import Module
from toydl.core.scalar import Scalar
from toydl.layer.linear import Linear


class ActivationType(StrEnum):
    RELU = "relu"
    SIGMOID = "sigmoid"


@dataclass
class MLPConfig:
    in_size: int
    out_size: int
    hidden_layer_num: int
    hidden_layer_size: int
    hidden_activation: ActivationType = ActivationType.RELU


class MLPBinaryClassifyNetFactory(Module):
    def __init__(self, config: MLPConfig):
        super().__init__()
        self.order_layer_names: list[str] = []
        self.config = self.__check_config(config)
        self.__construct_layers()

    @staticmethod
    def __check_config(config: MLPConfig):
        assert config.out_size == 1, "The out_size in MLPBinaryClassify should be 1"
        return config

    def __construct_layers(self):
        if self.config.hidden_layer_num == 0:
            setattr(
                self,
                "layer_input_output",
                Linear(self.config.in_size, self.config.out_size),
            )
            self.order_layer_names = ["layer_input_output"]
            return

        order_layer_names: list[str] = []
        # input layer
        setattr(
            self,
            "layer_input_hidden_0",
            Linear(self.config.in_size, self.config.hidden_layer_size),
        )
        order_layer_names.append("layer_input_hidden_0")
        # hidden layers: There is already one hidden layer between layer_input and layer_output, so we construct an additional hidden_layer_num - 1 hidden layers here.
        for layer_index in range(self.config.hidden_layer_num - 1):
            middle_layer_name = f"layer_hidden_{layer_index}_{layer_index + 1}"
            setattr(
                self,
                middle_layer_name,
                Linear(self.config.hidden_layer_size, self.config.hidden_layer_size),
            )
            order_layer_names.append(middle_layer_name)

        # output layers
        setattr(
            self,
            "layer_hidden_output",
            Linear(self.config.hidden_layer_size, self.config.out_size),
        )
        order_layer_names.append("layer_hidden_output")

        self.order_layer_names = order_layer_names

    def forward(self, xs: list[Scalar]) -> Scalar:
        for layer_name in self.order_layer_names:
            xs = getattr(self, layer_name).forward(xs)
            if "output" not in layer_name:
                if self.config.hidden_activation == ActivationType.RELU:
                    xs = [h.relu() for h in xs]
                elif self.config.hidden_activation == ActivationType.SIGMOID:
                    xs = [h.sigmoid() for h in xs]
        # note that our output size is 1
        output = xs[0].sigmoid()
        return output


if __name__ == "__main__":
    mlp_config = MLPConfig(
        in_size=2, out_size=1, hidden_layer_size=3, hidden_layer_num=1
    )
    mlp = MLPBinaryClassifyNetFactory(mlp_config)



from hypothesis import settings
from hypothesis.strategies import floats

import toydl

settings.register_profile("ci", deadline=None)
settings.load_profile("ci")


small_floats = floats(min_value=-100, max_value=100, allow_nan=False)


def assert_close(a, b):
    assert toydl.core.operator.is_close(a, b), "Failure x=%f y=%f" % (a, b)



from __future__ import annotations

import abc
from abc import abstractmethod
from typing import Any, Iterable, Optional, Tuple, Union

import toydl.core.operator as operators

from toydl.core.scalar.bp import backpropagate
from toydl.core.scalar.context import Context
from toydl.core.scalar.history import ScalarHistory

ScalarLike = Union[float, int, "Scalar"]


_var_count: int = 0


class Scalar:
    """
    A reimplementation of scalar values for auto-differentiation
    tracking. Scalar Variables behave as close as possible to standard
    Python numbers while also tracking the operations that led to the
    number's creation. They can only be manipulated by
    `ScalarFunction`.
    """

    def __init__(
        self,
        v: float,
        history: ScalarHistory | None = None,
        name: Optional[str] = None,
    ):
        global _var_count
        _var_count += 1
        self._unique_id: int = _var_count
        self.data: float = float(v)
        self.history = history
        self.derivative: Optional[float] = None
        if name is not None:
            self.name = name
        else:
            self.name = str(self.unique_id)

    def __repr__(self) -> str:
        return (
            f"Scalar(name={self.name}, unique_id={self.unique_id}, data={self.data:.4f}, "
            f"derivative={round(self.derivative, 4) if self.derivative else None})"
        )

    def __mul__(self, b: ScalarLike):
        return Mul.apply(self, b)

    def __rmul__(self, b: ScalarLike):
        return Mul.apply(b, self)

    def __truediv__(self, b: ScalarLike):
        return Mul.apply(self, Inv.apply(b))

    def __rtruediv__(self, b: ScalarLike):
        return Mul.apply(b, Inv.apply(self))

    def __add__(self, b: ScalarLike):
        return Add.apply(self, b)

    def __bool__(self):
        return bool(self.data)

    def __lt__(self, b: ScalarLike):
        return LT.apply(self, b)

    def __gt__(self, b: ScalarLike):
        return LT.apply(b, self)

    def __eq__(self, b: ScalarLike):  # type: ignore[override]
        return EQ.apply(self, b)

    def __sub__(self, b: ScalarLike):
        return Add.apply(self, Neg.apply(b))

    def __rsub__(self, b: ScalarLike):
        return Add.apply(b, Neg.apply(self))

    def __neg__(self):
        return Neg.apply(self)

    def log(self):
        return Log.apply(self)

    def exp(self):
        return Exp.apply(self)

    def sigmoid(self):
        return Sigmoid.apply(self)

    def relu(self):
        return ReLU.apply(self)

    @property
    def unique_id(self):
        return self._unique_id

    def requires_grad_(self, flag: bool = True):
        """
        Set the requires_grad flag to `flag` on variable.

        Ensures that operations on this variable will trigger
        backpropagation.

        :param flag: whether to require grad
        """
        if flag:
            self.history = ScalarHistory()

    def accumulate_derivative(self, x: Any) -> None:
        """
        Add `x` to the derivative accumulated on this variable.
        Should only be called during auto-differentiation on leaf variables.

        :param x: value to be accumulated
        """
        assert self.is_leaf(), "Only leaf variables can have derivatives."
        if self.derivative is None:
            self.derivative = 0.0
        self.derivative += x

    def is_leaf(self) -> bool:
        """True if this variable created by the user (no `last_fn`)"""
        return self.history is not None and self.history.last_fn is None

    def is_constant(self) -> bool:
        return self.history is None

    @property
    def parents(self) -> Iterable[Scalar]:
        assert self.history is not None
        return self.history.inputs

    def chain_rule(self, d_output: Any) -> Iterable[Tuple[Scalar, Any]]:
        h = self.history
        assert h is not None
        assert h.last_fn is not None
        assert h.ctx is not None

        derivatives = h.last_fn.backward(h.ctx, d_output)
        variables = h.inputs
        var_derivatives = [
            (var, derivative)
            for (var, derivative) in zip(variables, derivatives)
            if not var.is_constant()
        ]
        return var_derivatives

    def backward(self, d_output: Optional[float] = None) -> None:
        """
        Calls autodiff to fill in the derivatives for the history of this object.

        Args:
            d_output (number, opt): starting derivative to backpropagate through the model
                                   (typically left out, and assumed to be 1.0).
        """
        if d_output is None:
            d_output = 1.0
        backpropagate(self, d_output)


class ScalarFunction(abc.ABC):
    """
    A wrapper for a mathematical function that processes and produces
    Scalar variables.

    This is a static class and is never instantiated. We use `class`
    here to group together the `forward` and `backward` code.
    """

    @staticmethod
    @abstractmethod
    def _forward(*args: Any, **kwargs: Any) -> float:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _backward(ctx: Context, d_output: float) -> Union[float, Tuple[float, ...]]:
        raise NotImplementedError

    @classmethod
    def backward(cls, ctx: Context, d_out: float) -> Tuple[float, ...]:
        out = cls._backward(ctx, d_out)
        if isinstance(out, tuple):
            return out
        return (out,)

    @classmethod
    def forward(cls, ctx: Context, *inputs: float) -> float:
        return cls._forward(ctx, *inputs)  # type: ignore

    @classmethod
    def apply(cls, *vals: ScalarLike) -> Scalar:
        raw_vals: list[float] = []
        scalars: list[Scalar] = []
        for v in vals:
            if isinstance(v, Scalar):
                scalars.append(v)
                raw_vals.append(v.data)
            else:
                scalars.append(Scalar(v))
                raw_vals.append(v)

        # Call forward with the variables.
        ctx = Context()
        c = cls.forward(ctx, *raw_vals)
        assert isinstance(c, (float, int)), (
            f"Expected return type float or int got {type(c)}"
        )

        # Create a new variable from the result with a new history.
        history = ScalarHistory(cls, ctx, scalars)  # type: ignore
        return Scalar(c, history)


class Add(ScalarFunction):
    """Addition function :math:`f(x, y) = x + y`"""

    @staticmethod
    def _forward(ctx: Context, a: float, b: float) -> float:
        return a + b

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> Tuple[float, ...]:
        return d_output, d_output


class Log(ScalarFunction):
    """Log function :math:`f(x) = log(x)`"""

    @staticmethod
    def _forward(ctx: Context, a: float) -> float:
        ctx.save_for_backward(a)
        return operators.log(a)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> float:
        a = ctx.saved_values[0]
        return operators.log_back(a, d_output)


class Mul(ScalarFunction):
    """Multiplication function"""

    @staticmethod
    def _forward(ctx: Context, a: float, b: float) -> float:
        ctx.save_for_backward(a, b)
        return operators.mul(a, b)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> tuple[float, float]:
        a, b = ctx.saved_values
        return operators.mul_back(a, b, d_output)


class Inv(ScalarFunction):
    """Inverse function"""

    @staticmethod
    def _forward(ctx: Context, a: float) -> float:
        ctx.save_for_backward(a)
        return operators.inv(a)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> float:
        a = ctx.saved_values[0]
        return operators.inv_back(a, d_output)


class Neg(ScalarFunction):
    """Negation function"""

    @staticmethod
    def _forward(ctx: Context, a: float) -> float:
        return operators.neg(a)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> float:
        return operators.neg_back(d_output)


class Sigmoid(ScalarFunction):
    """Sigmoid function"""

    @staticmethod
    def _forward(ctx: Context, a: float) -> float:
        ctx.save_for_backward(a)
        return operators.sigmoid(a)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> float:
        a = ctx.saved_values[0]
        return operators.sigmoid_back(a, d_output)


class ReLU(ScalarFunction):
    """ReLU function"""

    @staticmethod
    def _forward(ctx: Context, a: float) -> float:
        ctx.save_for_backward(a)
        return operators.relu(a)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> float:
        a = ctx.saved_values[0]
        return operators.relu_back(a, d_output)


class Exp(ScalarFunction):
    """Exp function"""

    @staticmethod
    def _forward(ctx: Context, a: float) -> float:
        ctx.save_for_backward(a)
        return operators.exp(a)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> float:
        a = ctx.saved_values[0]
        return operators.exp_back(a, d_output)


class LT(ScalarFunction):
    """Less-than function :math:`f(x) =` 1.0 if x is less than y else 0.0"""

    @staticmethod
    def _forward(ctx: Context, a: float, b: float) -> float:
        return operators.lt(a, b)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> Tuple[float, float]:
        return 0.0, 0.0


class EQ(ScalarFunction):
    """Equal function :math:`f(x) =` 1.0 if x is equal to y else 0.0"""

    @staticmethod
    def _forward(ctx: Context, a: float, b: float) -> float:
        return operators.eq(a, b)

    @staticmethod
    def _backward(ctx: Context, d_output: float) -> Tuple[float, float]:
        return 0.0, 0.0



import math


def mul(x: float, y: float) -> float:
    """$f(x, y) = x * y$"""
    return x * y


def mul_back(x: float, y: float, d: float):
    return d * y, d * x


def id_(x: float) -> float:
    """$f(x) = x$"""
    return x


def add(x: float, y: float) -> float:
    """$f(x, y) = x + y$"""
    return x + y


def neg(x: float) -> float:
    """$f(x) = -x$"""
    return -x


def neg_back(d: float) -> float:
    return -d


def lt(x: float, y: float) -> float:
    """$f(x, y) = x < y$"""
    return float(x < y)


def eq(x: float, y: float) -> float:
    return float(abs(x - y) <= 1e-8)


def max_(x: float, y: float) -> float:
    return x if x > y else y


def is_close(x: float, y: float) -> float:
    return float(abs(x - y) < 1e-2)


def sigmoid(x: float) -> float:
    r"""
    $$f(x) =  \frac{1.0}{(1.0 + e^{-x})}$$

    Calculate as

    $$
    f(x) = \begin{cases}
    \frac{1.0}{1.0 + e^{-x}} & \text{if } x \geq 0 \\
    \frac{e^x}{1.0 + e^x} & \text{otherwise}
    \end{cases}
    $$

    for stability.
    The key is to make sure the `x` in exp(x) is always negative to avoid exp(x) overflow.
    """
    return 1 / (1 + math.exp(-x)) if x >= 0 else math.exp(x) / (1 + math.exp(x))


def sigmoid_back(x: float, d: float) -> float:
    return d * sigmoid(x) * (1 - sigmoid(x))


def relu(x: float) -> float:
    """
    f(x) = x if x is greater than 0, else 0
    """
    return x if x > 0 else 0


def log(x: float) -> float:
    return math.log(x + 1e-6)


def exp(x: float) -> float:
    return math.exp(x)


def exp_back(x: float, d: float) -> float:
    return d * exp(x)


def log_back(x: float, d: float) -> float:
    return d / (x + 1e-6)


def inv(x: float) -> float:
    return 1 / x


def inv_back(x: float, d: float) -> float:
    return -d / x**2


def relu_back(x: float, d: float) -> float:
    return d if x > 0 else 0



from pathlib import Path
from typing import List, Tuple

import matplotlib.pyplot as plt

import toydl.dataset.simulation as simulation_dataset

from toydl.core.optim import SGD, Momentum, Optimizer
from toydl.core.scalar.scalar import Scalar
from toydl.dataset.simple import SimpleDataset
from toydl.loss.cross_entropy import CrossEntropyLoss
from toydl.network.mlp import MLPBinaryClassifyNetFactory, MLPConfig


# --8<-- [start:model]
class MLPBinaryClassifyModel:
    def __init__(self, mlp_config: MLPConfig):
        self.net = MLPBinaryClassifyNetFactory(mlp_config)

    def forward_once(self, x: list[float], y: int) -> tuple[Scalar, Scalar]:
        y_pred = self.net.forward(list(Scalar(v) for v in x))
        loss = CrossEntropyLoss().forward(y_true=y, y_pred=y_pred)
        return y_pred, loss

    def evaluate(self, dateset: SimpleDataset) -> tuple[float, int]:
        # switch to eval mode
        self.net.eval()
        total_loss = 0.0
        correct = 0
        for x, y in dateset:
            y_pred, loss = self.forward_once(x, y)
            if y == 1:
                correct += 1 if y_pred.data > 0.5 else 0
            else:
                correct += 1 if y_pred.data < 0.5 else 0
            total_loss += loss.data

        # switch back to train mode
        self.net.train()

        return total_loss, correct

    def train(
        self,
        training_set: SimpleDataset,
        test_set: SimpleDataset,
        optimizer: Optimizer,
        max_epochs: int = 500,
    ) -> Tuple[List[float], List[float], str]:
        training_loss, testing_loss, test_correct = [], [], 0
        for epoch in range(1, max_epochs + 1):
            optimizer.zero_grad()

            # Forward & Backward
            for x, y in training_set:
                _, loss = self.forward_once(x, y)
                (loss / len(training_set)).backward()

            # Update parameters
            optimizer.step()

            # Evaluation
            train_loss, train_correct = self.evaluate(training_set)
            test_loss, test_correct = self.evaluate(test_set)

            training_loss.append(train_loss)
            testing_loss.append(test_loss)
            if epoch % 10 == 0 or epoch == max_epochs:
                print(
                    f"[Epoch {epoch}]Train Loss = {train_loss}, "
                    f"right({train_correct})/total({len(training_set)}) = {train_correct / len(training_set)}\n"
                    f"[Epoch {epoch}]Test  Loss = {test_loss},  "
                    f"right({test_correct})/total({len(test_set)}) = {test_correct / len(test_set)}"
                )
        test_result = f"right/total = {test_correct}/{len(test_set)}"
        return training_loss, testing_loss, test_result

    # --8<-- [end:model]


# --8<-- [start:gen_dateset]
def get_dataset(n: int = 100) -> Tuple[SimpleDataset, SimpleDataset]:
    data = simulation_dataset.diag(n)
    training_set, test_set = data.train_test_split(train_proportion=0.7)

    return training_set, test_set


# --8<-- [end:gen_dateset]


def run_sgd(
    mlp_config: MLPConfig,
    training_set: SimpleDataset,
    test_set: SimpleDataset,
    learning_rate: float,
    max_epochs: int = 500,
):
    mlp_model = MLPBinaryClassifyModel(mlp_config)

    sgd_optimizer = SGD(mlp_model.net.parameters(), learning_rate)

    training_loss, testing_loss, test_result = mlp_model.train(
        training_set, test_set, sgd_optimizer, max_epochs=max_epochs
    )
    return training_loss, testing_loss, test_result


def run_momentum(
    mlp_config: MLPConfig,
    training_set: SimpleDataset,
    test_set: SimpleDataset,
    learning_rate: float,
    max_epochs: int = 500,
):
    momentum = 0.5
    mlp_model = MLPBinaryClassifyModel(mlp_config)

    optimizer = Momentum(mlp_model.net.parameters(), learning_rate, momentum)

    training_loss, testing_loss, test_result = mlp_model.train(
        training_set, test_set, optimizer, max_epochs=max_epochs
    )
    return training_loss, testing_loss, test_result


def plot_multiple_optimizers(
    optimizer_results: dict[str, tuple[list[float], list[float], str]],
    plot_type: str = "both",  # "train", "test", or "both"
    title: str = "Optimizer Comparison",
    filename: str = "optimizer_comparison.png",
    output_dir: Path | None = None,
):
    """
    Plot losses for multiple optimizers in one figure for comparison.
    """
    plt.clf()
    plt.figure(figsize=(10, 6))

    colors = ["r", "g", "b", "c", "m", "y", "k"]
    markers = ["o", "s", "^", "*", "x", "D", "v"]

    color_idx = 0
    for optimizer_name, (training_loss, testing_loss, _) in optimizer_results.items():
        color = colors[color_idx % len(colors)]
        marker = markers[color_idx % len(markers)]

        if plot_type in ["train", "both"]:
            plt.plot(
                training_loss,
                f"{color}-",
                label=f"{optimizer_name} - Training",
                marker=marker,
                markevery=max(1, len(training_loss) // 10),
            )

        if plot_type in ["test", "both"]:
            plt.plot(
                testing_loss,
                f"{color}--",
                label=f"{optimizer_name} - Test",
                marker=marker,
                markevery=max(1, len(testing_loss) // 10),
            )

        color_idx += 1

    plt.title(title)
    plt.xlabel("Epoch")
    plt.ylabel("Loss")
    plt.legend()
    plt.grid(True, linestyle="--", alpha=0.7)
    plt.tight_layout()

    # Handle output directory
    if output_dir is not None:
        output_dir.mkdir(exist_ok=True, parents=True)
        save_path = output_dir / filename
    else:
        save_path = Path(f"./{filename}")

    plt.savefig(save_path, dpi=300)
    plt.show()

    # Print the test results
    print("\nTest Results:")
    for optimizer_name, (_, _, test_result) in optimizer_results.items():
        print(f"{optimizer_name}: {test_result}")


def run():
    # Setup output directory using pathlib
    current_file = Path(__file__)
    image_dir = current_file.parent / "images"
    image_dir.mkdir(exist_ok=True, parents=True)

    n = 100
    training_set, test_set = get_dataset(n)
    training_set.plot(filename=str(image_dir / "training_set.png"))
    mlp_config = MLPConfig(
        in_size=2, out_size=1, hidden_layer_size=10, hidden_layer_num=2
    )

    learning_rate = 0.01
    max_epochs = 500

    sgd_training_loss, sgd_testing_loss, sgd_test_result = run_sgd(
        mlp_config, training_set, test_set, learning_rate, max_epochs
    )
    momentum_training_loss, momentum_testing_loss, momentum_result = run_momentum(
        mlp_config, training_set, test_set, learning_rate, max_epochs
    )

    # Plot comparison of optimizers
    optimizer_results = {
        "SGD": (sgd_training_loss, sgd_testing_loss, sgd_test_result),
        "Momentum": (momentum_training_loss, momentum_testing_loss, momentum_result),
    }
    plot_multiple_optimizers(
        optimizer_results,
        title="Optimizer Loss Comparison",
        filename="optimizer_comparison.png",
        output_dir=image_dir,
    )


if __name__ == "__main__":
    run()



from ._base import Optimizer

from typing import Sequence

from toydl.core.module import Parameter


class Momentum(Optimizer):
    """
    Momentum Optimizer
    """

    def __init__(
        self, parameters: Sequence[Parameter], lr: float = 0.01, momentum: float = 0.9
    ):
        """
        Init the Momentum optimizer

        :param parameters: the parameters that will be optimized
        :param lr: learning rate
        :param momentum: momentum coefficient
        """
        super().__init__(parameters)
        self.lr = lr
        self.momentum = momentum
        self.parameter_delta_map: dict[Parameter, float] = {}

    def zero_grad(self) -> None:
        """
        Clear the grad/derivative value of parameter
        """
        for p in self.parameters:
            if hasattr(p.value, "derivative") and p.value.derivative is not None:
                p.value.derivative = None

    def step(self) -> None:
        """
        Run a momentum step to update parameter value
        """
        for p in self.parameters:
            if hasattr(p.value, "derivative") and p.value.derivative is not None:
                delta = (
                    -self.lr * p.value.derivative
                    + self.momentum * self.parameter_delta_map.get(p, 0)
                )
                self.parameter_delta_map[p] = delta
                new_value = p.value + delta
                p.update(new_value)



import math

import pytest

from hypothesis import given

from toydl.core.operator import (
    add,
    eq,
    id_,
    inv,
    inv_back,
    log_back,
    lt,
    max_,
    mul,
    neg,
    relu,
    relu_back,
    sigmoid,
)

from .strategies import assert_close, small_floats


@pytest.mark.operator
@given(small_floats, small_floats)
def test_same_as_python(x, y):
    """Check that the main operators all return the same value of the python version"""
    assert_close(mul(x, y), x * y)
    assert_close(add(x, y), x + y)
    assert_close(neg(x), -x)
    assert_close(max_(x, y), x if x > y else y)
    if x != 0.0 and not math.isinf(1.0 / x):
        assert_close(inv(x), 1.0 / x)


@pytest.mark.operator
@given(small_floats)
def test_relu(a):
    if a > 0:
        assert relu(a) == a
    if a < 0:
        assert relu(a) == 0.0


@pytest.mark.operator
@given(small_floats, small_floats)
def test_relu_back(a, b):
    if a > 0:
        assert relu_back(a, b) == b
    if a < 0:
        assert relu_back(a, b) == 0.0


@pytest.mark.operator
@given(small_floats)
def test_id(a):
    assert id_(a) == a


@pytest.mark.operator
@given(small_floats)
def test_lt(a):
    """Check that a - 1.0 is always less than a"""
    assert lt(a - 1.0, a) == 1.0
    assert lt(a, a - 1.0) == 0.0


@pytest.mark.operator
@given(small_floats)
def test_max(a):
    assert max_(a - 1.0, a) == a
    assert max_(a, a - 1.0) == a
    assert max_(a + 1.0, a) == a + 1.0
    assert max_(a, a + 1.0) == a + 1.0


@pytest.mark.operator
@given(small_floats)
def test_eq(a):
    assert eq(a, a) == 1.0
    assert eq(a, a - 1.0) == 0.0
    assert eq(a, a + 1.0) == 0.0


# ## Task 0.2 - Property Testing

# Implement the following property checks
# that ensure that your operators obey basic
# mathematical rules.


@pytest.mark.operator
@given(small_floats)
def test_sigmoid(a):
    """Check properties of the sigmoid function, specifically
    * It is always between 0.0 and 1.0.
    * one minus sigmoid is the same as negative sigmoid
    * It crosses 0 at 0.5
    * it is  strictly increasing.
    """
    ans = sigmoid(a)
    assert 0 <= ans <= 1
    assert_close(1 - ans, sigmoid(-a))
    assert (a >= 0 and ans >= 0.5) or (a <= 0 and ans <= 0.5)
    assert sigmoid(a + 0.1) >= ans


@pytest.mark.operator
@given(small_floats, small_floats, small_floats)
def test_transitive(a, b, c):
    """Test the transitive property of less-than (a < b and b < c implies a < c)"""
    if (a < b) and (b < c):
        assert a < c


@pytest.mark.operator
@given(small_floats, small_floats)
def test_symmetric(a, b):
    """
    Write a test that ensures that mul is symmetric, i.e.
    gives the same value regardless of the order of its input.
    """
    assert_close(mul(a, b), mul(b, a))


@pytest.mark.operator
@given(small_floats, small_floats, small_floats)
def test_distribute(x, y, z):
    r"""
    Write a test that ensures that your operators distribute, i.e.
    :math:`z \times (x + y) = z \times x + z \times y`
    """
    assert_close(mul(z, (x + y)), mul(z, x) + mul(z, y))


@pytest.mark.operator
@given(small_floats, small_floats)
def test_backs(a, b):
    relu_back(a, b)
    inv_back(a + 2.4, b)
    log_back(abs(a) + 4, b)



from dataclasses import dataclass, field

from toydl.core.scalar import Scalar
from toydl.loss.base import LossBase


@dataclass(frozen=True, repr=True)
class CrossEntropyLoss(LossBase):
    name: str = field(default="Cross Entropy Loss")

    @staticmethod
    def forward(y_true: int, y_pred: Scalar) -> Scalar:
        return -1 * ((y_true * y_pred.log()) + (1 - y_true) * ((1 - y_pred).log()))



import random

from toydl.core.module import Module, Parameter
from toydl.core.scalar import Scalar


class Linear(Module):
    def __init__(self, in_size: int, out_size: int):
        super().__init__()
        self.weights: list[list[Parameter]] = []  # type: ignore[var-annotated]
        self.bias: list[Parameter] = []
        for i in range(in_size):
            self.weights.append([])
            for j in range(out_size):
                self.weights[i].append(
                    self.add_parameter(
                        f"weight_{i}_{j}", Scalar(2 * (random.random() - 0.5))
                    )
                )
        for j in range(out_size):
            self.bias.append(
                self.add_parameter(f"bias_{j}", Scalar(2 * (random.random() - 0.5)))
            )

    def forward(self, xs: list[Scalar]) -> list[Scalar]:
        outputs: list[Scalar] = []
        n, m = len(self.weights), len(self.weights[0])
        for j in range(m):
            output = Scalar(0)
            for i in range(n):
                output += self.weights[i][j].value * xs[i] + self.bias[j].value
            outputs.append(output)
        return outputs



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, Optional

if TYPE_CHECKING:
    from toydl.core.scalar.scalar import Scalar, ScalarFunction

from dataclasses import dataclass

from toydl.core.scalar.context import Context


@dataclass
class ScalarHistory:
    """
    `ScalarHistory` stores the history of `Function` operations that was
    used to construct the current Variable.

    Attributes:
        last_fn : The last Function that was called.
        ctx : The context.py for that Function.
        inputs : The inputs that were given when `last_fn.forward` was called.

    """

    last_fn: Optional["ScalarFunction"] = None
    ctx: Context | None = None
    inputs: Sequence["Scalar"] = ()



import random

from typing import List

from toydl.dataset.simple import SimpleDataset


def generate_random_points(n: int) -> List[List[float]]:
    points = [[random.random(), random.random()] for _ in range(n)]
    return points


def simple(n: int) -> SimpleDataset:
    xs = generate_random_points(n)
    ys = []
    for x1, x2 in xs:
        y = 1 if x1 < 0.5 else 0
        ys.append(y)
    return SimpleDataset(xs, ys)


def diag(n: int) -> SimpleDataset:
    xs = generate_random_points(n)
    ys = []
    for x1, x2 in xs:
        y1 = 1 if x1 + x2 < 0.5 else 0
        ys.append(y1)
    return SimpleDataset(xs, ys)


def split(n: int) -> SimpleDataset:
    xs = generate_random_points(n)
    ys = []
    for x1, x2 in xs:
        y = 1 if x1 < 0.2 or x1 > 0.8 else 0
        ys.append(y)
    return SimpleDataset(xs, ys)


def xor(n: int) -> SimpleDataset:
    xs = generate_random_points(n)
    ys = []
    for x1, x2 in xs:
        y = 1 if ((x1 < 0.5 < x2) or (x1 > 0.5 > x2)) else 0
        ys.append(y)
    return SimpleDataset(xs, ys)



from dataclasses import dataclass
from typing import Any


@dataclass
class Context:
    """
    Context class is used by `ScalarFunction` to store information during the forward pass.
    """

    no_grad: bool = False
    saved_values: tuple[Any, ...] = ()

    def save_for_backward(self, *values: Any) -> None:
        """Store the given `values` if they need to be used during backpropagation.

        :param values: The values that should be saved for backward
        """
        if self.no_grad:
            return
        self.saved_values = values



import marimo

__generated_with = "0.14.13"
app = marimo.App(width="full", layout_file="layouts/backword.slides.json")


@app.cell(hide_code=True)
def _():
    import marimo as mo

    return (mo,)


@app.cell
def _():
    import toydl
    import math

    from toydl.core.scalar import Scalar
    from toydl.core.scalar.bp import topological_sort
    from toydl.network.mlp import MLPConfig, MLPBinaryClassifyNetFactory, ActivationType

    return (
        ActivationType,
        MLPBinaryClassifyNetFactory,
        MLPConfig,
        Scalar,
        math,
        topological_sort,
        toydl,
    )


@app.cell
def _(toydl):
    toydl.__version__
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Simple function backward""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    $$ y = f(x_1, x_2) = 3x_1 + 5x_2 $$

    $$ f'(x_1) = 3 $$

    $$ f'(x_2)  = 5 $$
    """
    )
    return


@app.cell
def _():
    return


@app.cell
def _(Scalar):
    x1 = Scalar(1.0, name="x1")
    x1.requires_grad_()
    x2 = Scalar(2.0, name="x2")
    x2.requires_grad_()

    y = Scalar(3, name="x1_coef") * x1 + Scalar(5, name="x2_coef") * x2
    y.name = "y"

    y
    return x1, x2, y


@app.cell
def _(topological_sort, y):
    topological_sort(y)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""注意到，除了我们指定的一些 `Scalar` 之外，这里存在一些存储中间计算结果的 `Scalar`."""
    )
    return


@app.cell
def _(math, x1, x2, y):
    y.backward(d_output=1.0)

    assert math.isclose(x1.derivative, 3.0)
    assert math.isclose(x2.derivative, 5.0)
    return


@app.cell
def _(x1):
    x1
    return


@app.cell
def _(x2):
    x2
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## MLP backward""")
    return


@app.cell
def _(ActivationType, MLPBinaryClassifyNetFactory, MLPConfig):
    mlp_config = MLPConfig(
        in_size=2,
        out_size=1,
        hidden_layer_size=3,
        hidden_layer_num=1,
        hidden_activation=ActivationType.SIGMOID,
    )
    mlp = MLPBinaryClassifyNetFactory(mlp_config)
    return (mlp,)


@app.cell
def _(mlp):
    for _p in mlp.named_parameters():
        print(_p)
    return


@app.cell
def _(Scalar, mlp):
    x1_1 = Scalar(1.0, name="x1")
    x2_1 = Scalar(2.0, name="x2")
    xs = [x1_1, x2_1]
    y_1 = mlp.forward(xs)
    y_1.name = "y"
    y_1
    return x1_1, x2_1, y_1


@app.cell
def _(y_1):
    y_1.backward(d_output=1.0)
    return


@app.cell
def _(mlp):
    for _p in mlp.named_parameters():
        print(_p)
    return


@app.cell
def _(x1_1):
    x1_1
    return


@app.cell
def _(x2_1):
    x2_1
    return


if __name__ == "__main__":
    app.run()



from dataclasses import dataclass
from typing import List, Tuple, Union

import matplotlib.pyplot as plt

from toydl.dataset.base import DataSetBase
from toydl.dataset.exception import DatasetValidationError


@dataclass(repr=True)
class SimpleDataset(DataSetBase):
    xs: List[List[Union[int, float]]]
    ys: List[int]

    def __post_init__(self):
        if len(self.xs) != len(self.ys):
            raise DatasetValidationError(
                f"X, y lengths don't match, "
                f"lengths: X({len(self.xs)}), y({len(self.ys)})"
            )

        item_lengths = [len(xi) for xi in self.xs]
        if len(set(item_lengths)) != 1:
            raise DatasetValidationError(
                f"X items have different lengths: {set(item_lengths)}"
            )

    def __len__(self) -> int:
        return len(self.xs)

    def __getitem__(self, index) -> "SimpleDataset":
        if isinstance(index, slice):
            return SimpleDataset(self.xs[index], self.ys[index])
        elif isinstance(index, int):
            return SimpleDataset([self.xs[index]], [self.ys[index]])
        else:
            raise TypeError(f"Invalid argument type: {type(index)}, {index}")

    def __iter__(self):
        return iter(zip(self.xs, self.ys))

    def train_test_split(
        self, train_proportion: float = 0.7
    ) -> Tuple["SimpleDataset", "SimpleDataset"]:
        training_size = int(len(self) * train_proportion)
        training_set = self[:training_size]
        test_set = self[training_size:]
        return training_set, test_set

    def plot(self, filename: str = "dataset.png"):
        x_values = [x[0] for x in self.xs]
        y_values = [x[1] for x in self.xs]

        # Assign colors based on the y values
        colors = ["green" if y == 0 else "red" for y in self.ys]

        # Plot the data points with specified colors
        plt.scatter(x_values, y_values, c=colors)
        plt.tight_layout()
        if filename is not None:
            plt.savefig(filename, dpi=300)
        # plt.show()


if __name__ == "__main__":
    data = SimpleDataset([[1, 1], [2, 2]], [1, 2])
    data.plot()



from typing import Any

from toydl.core.scalar import Scalar


class Parameter:
    """
    A Parameter is a special container stored in a :class:`Module`.

    It is designed to hold a :class:`Variable`, but we allow it to hold
    any value for testing.
    """

    def __init__(self, value: Scalar, name: str | None = None):
        """
        :param value: the value of parameter
        :param name: the name of parameter
        """
        self.value = value
        self.name = name
        # Set `requires_grad_` to True, indicating that these parameters are trainable (can accumulate gradients)
        self.value.requires_grad_(True)
        if self.name:
            self.value.name = self.name

    def update(self, x: Scalar) -> None:
        r"""Update the parameter value.

        :param x: the parameter's new value
        """
        self.value.data = x.data

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return str(self.value)


class Module:
    """
    Modules form a tree that stores parameters and other submodules.
    They make up the basis of neural network stacks.

    """

    def __init__(self):
        self._modules: dict[str, "Module"] = {}
        self._parameters: dict[str, Parameter] = {}
        self.training = True

    def modules(self):
        """Return the direct child modules of this module."""
        return self.__dict__["_modules"].values()

    def train(self):
        """Set the mode of this module and all descendant modules to `train`."""
        self.training = True
        for m in self.modules():
            m.training = True

    def eval(self):
        """Set the mode of this module and all descendant modules to `eval`."""
        self.training = False
        for m in self.modules():
            m.training = False

    def named_parameters(self) -> list[tuple[str, Parameter]]:
        """
        Collect all the parameters of this module and its descendants.


        :return list of pairs: Contains the name and :class:`Parameter` of each ancestor parameter.
        """
        named_params: list[tuple[str, Parameter]] = []
        # the module params
        for name, param in self._parameters.items():
            named_params.append((name, param))
        # descendants params
        for module_name, module in self._modules.items():
            for param_name, param in module.named_parameters():
                named_params.append((f"{module_name}.{param_name}", param))
        return named_params

    def parameters(self):
        """Enumerate over all the parameters of this module and its descendants."""
        params = [param for _, param in self.named_parameters()]
        return params

    def add_parameter(self, k: str, v: Scalar):
        """
        Manually add a parameter. Useful helper for scalar parameters.

        :param k: Local name of the parameter.
        :param v: Value for the parameter.
        :return parameter: Newly created parameter.
        """
        val = Parameter(v, k)
        self.__dict__["_parameters"][k] = val
        return val

    def __setattr__(self, key: Any, val: Any) -> None:
        if isinstance(val, Parameter):
            self.__dict__["_parameters"][key] = val
        elif isinstance(val, Module):
            self.__dict__["_modules"][key] = val
        else:
            super().__setattr__(key, val)

    def __getattr__(self, key: Any) -> Any:
        if key in self.__dict__["_parameters"]:
            return self.__dict__["_parameters"][key]

        if key in self.__dict__["_modules"]:
            return self.__dict__["_modules"][key]

        return super().__getattribute__(key)

    def __call__(self, xs: list[Scalar]) -> list[Scalar] | Scalar:
        return self.forward(xs)

    def forward(self, xs: list[Scalar]) -> list[Scalar] | Scalar:
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable

if TYPE_CHECKING:
    from toydl.core.scalar.scalar import Scalar


def topological_sort(variable: "Scalar") -> Iterable["Scalar"]:
    """
    Computes the topological order of the computation graph.

    :param variable: The right-most variable
    :return: Non-constant Variables in topological order starting from the right.
    """
    variables: list["Scalar"] = []
    visited: set[int] = set()

    def visit(var: "Scalar"):
        if var.is_constant():
            return
        if var.unique_id in visited:
            return
        visited.add(var.unique_id)
        if var.parents:
            for p in var.parents:
                visit(p)
        variables.append(var)

    visit(variable)
    return variables[::-1]


def backpropagate(variable: "Scalar", deriv: Any) -> None:
    """
    Runs backpropagation on the computation graph in order to
    compute derivatives for the leave nodes.

    :param variable: The right-most variable
    :param deriv: Its derivative that we want to propagate backward to the leaves.

    :return: No return.
    Should write to its results to the derivative values of each leaf through `accumulate_derivative`.
    """
    sorted_values = topological_sort(variable)
    var_derivative_map = {variable.unique_id: deriv}
    for var in sorted_values:
        if var.is_leaf():
            var.accumulate_derivative(var_derivative_map[var.unique_id])
        else:
            var_derivatives = var.chain_rule(var_derivative_map[var.unique_id])
            for _var, _derivative in var_derivatives:
                var_derivative_map[_var.unique_id] = (
                    var_derivative_map.get(_var.unique_id, 0) + _derivative
                )
